# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_argument_iterable,
    check_duration,
    check_numeric_numpy_array,
    check_operator,
    check_sample_times,
)

from .documentation import Category
from .node_data import (
    ConvolutionKernel,
    Pwc,
    Stf,
    Tensor,
)
from .types import TensorLike
from .utils import (
    validate_batch_and_value_shapes,
    validate_hamiltonian,
    validate_shape,
)


class StfOperator(Node):
    r"""
    Create a constant operator multiplied by a sampleable signal.

    Parameters
    ----------
    signal : Stf
        A sampleable function representing the signal :math:`a(t)`
        or a batch of sampleable functions.
    operator : np.ndarray or Tensor
        The operator :math:`A`. It must have two equal dimensions.

    Returns
    -------
    Stf
        The sampleable operator :math:`a(t)A` (or batch of sampleable operators, if
        you provide a batch of signals).

    See Also
    --------
    Graph.constant_stf_operator : Create a constant `Stf` operator.
    Graph.pwc_operator : Corresponding operation for `Pwc`\s.
    Graph.stf_sum : Sum multiple `Stf`\s.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "stf_operator"
    args = [forge.arg("signal", type=Stf), forge.arg("operator", type=TensorLike)]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        signal = kwargs.get("signal")
        operator = kwargs.get("operator")
        check_argument(
            isinstance(signal, Stf), "The signal must be an Stf.", {"signal": signal}
        )
        batch_shape, _ = validate_batch_and_value_shapes(signal, "signal")
        value_shape = validate_shape(operator, "operator")
        check_operator(operator, "operator")
        check_argument(
            len(value_shape) == 2,
            "The operator must be a matrix, not a batch.",
            {"operator": operator},
            extras={"operator.shape": value_shape},
        )
        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)


class ConstantStfOperator(Node):
    r"""
    Create a constant operator.

    Parameters
    ----------
    operator : np.ndarray or Tensor
        The operator :math:`A`, or a batch of operators. It must have at
        least two dimensions, and its last two dimensions must be equal.

    Returns
    -------
    Stf(3D)
        The operator :math:`t\mapsto A` (or batch of
        operators, if you provide a batch of operators).

    See Also
    --------
    Graph.constant_pwc_operator : Corresponding operation for `Pwc`\s.
    Graph.constant_stf : Create a batch of constant `Stf`\s.
    Graph.stf_operator : Create an `Stf` operator.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "constant_stf_operator"
    args = [forge.arg("operator", type=TensorLike)]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        operator = kwargs.get("operator")
        shape = validate_shape(operator, "operator")
        check_operator(operator, "operator")
        return Stf(_operation, value_shape=shape[-2:], batch_shape=shape[:-2])


class ConstantStf(Node):
    r"""
    Create a constant sampleable tensor-valued function of time.

    Parameters
    ----------
    constant : number or np.ndarray or Tensor
        The constant value :math:`c` of the function.
        To create a batch of :math:`B_1 \times \ldots \times B_n` constant
        functions of shape :math:`D_1 \times \ldots \times D_m`, provide this `constant`
        parameter as an object of shape
        :math:`B_1\times\ldots\times B_n\times D_1\times\ldots\times D_m`.
    batch_dimension_count : int, optional
        The number of batch dimensions, :math:`n`, in `constant`.
        If provided, the first :math:`n` dimensions of `constant` are considered batch dimensions.
        Defaults to 0, which corresponds to no batch.

    Returns
    -------
    Stf
       An Stf representing the constant function :math:`f(t) = c` for all time
       :math:`t` (or a batch of functions, if you provide `batch_dimension_count`).

    See Also
    --------
    Graph.constant_pwc : Corresponding operation for `Pwc`\s.
    Graph.constant_stf_operator : Create a constant sampleable function from operators.
    Graph.identity_stf : Create an `Stf` representing the identity function.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "constant_stf"
    args = [
        forge.arg("constant", type=Union[float, complex, np.ndarray, Tensor]),
        forge.arg("batch_dimension_count", type=int, default=0),
    ]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        constant = kwargs.get("constant")
        batch_dimension_count = kwargs.get("batch_dimension_count")

        check_numeric_numpy_array(constant, "constant")
        check_argument_integer(batch_dimension_count, "batch_dimension_count")

        shape = validate_shape(constant, "constant")
        check_argument(
            len(shape) >= batch_dimension_count,
            "The number of batch dimensions must not be larger than the number "
            "of dimensions of the input constant.",
            {"constant": constant, "batch_dimension_count": batch_dimension_count},
            extras={"Number of value dimensions": len(shape)},
        )
        check_argument(
            batch_dimension_count >= 0,
            "The number of batch dimensions must not be negative.",
            {"batch_dimension_count": batch_dimension_count},
        )

        return Stf(
            operation=_operation,
            value_shape=shape[batch_dimension_count:],
            batch_shape=shape[:batch_dimension_count],
        )


class StfSum(Node):
    r"""
    Create the sum of multiple sampleable functions.

    Parameters
    ----------
    terms : list[Stf]
        The individual sampleable function :math:`\{v_j(t)\}` to sum.

    Returns
    -------
    Stf
        The sampleable function of time :math:`\sum_j v_j(t)`. It has the same
        shape as each of the `terms` that you provide.

    See Also
    --------
    Graph.hermitian_part : Hermitian part of an `Stf` operator.
    Graph.pwc_sum : Corresponding operation for `Pwc`\s.
    Graph.stf_operator : Create an `Stf` operator.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "stf_sum"
    args = [forge.arg("terms", type=List[Stf])]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        terms = kwargs.get("terms")
        check_argument_iterable(terms, "terms")
        check_argument(
            all(isinstance(term, Stf) for term in terms),
            "Each of the terms must be an Stf.",
            {"terms": terms},
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(terms[0], "terms[0]")
        check_argument(
            all(
                (
                    (value_shape == term.value_shape)
                    and (batch_shape == term.batch_shape)
                )
                for term in terms[1:]
            ),
            "All the terms must have the same shape.",
            {"terms": terms},
        )
        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)


class DiscretizeStf(Node):
    r"""
    Create a piecewise-constant function by discretizing a sampleable function.

    Use this function to create a piecewise-constant approximation to a sampleable
    function (obtained, for example, by filtering an initial
    piecewise-constant function).

    Parameters
    ----------
    stf : Stf
        The sampleable function :math:`v(t)` to discretize. The values of the
        function can have any shape. You can also provide a batch of
        functions, in which case the discretization is applied to each
        element of the batch.
    duration : float
        The duration :math:`\tau` over which discretization should be
        performed. The resulting piecewise-constant function has this
        duration.
    segment_count : int
        The number of segments :math:`N` in the resulting piecewise-constant
        function.
    sample_count_per_segment : int, optional
        The number of samples :math:`M` of the sampleable function to take when
        calculating the value of each segment in the discretization. Defaults
        to 1.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc
        The piecewise-constant function :math:`w(t)` obtained by discretizing
        the sampleable function (or batch of piecewise-constant functions, if
        you provided a batch of sampleable functions).

    See Also
    --------
    Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
    Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
    Graph.identity_stf : Create an `Stf` representing the identity function.
    Graph.sample_stf : Sample an `Stf` at given times.

    Notes
    -----
    The resulting function :math:`w(t)` is piecewise-constant with :math:`N`
    segments, meaning it has segment values :math:`\{w_n\}` such that
    :math:`w(t)=w_n` for :math:`t_{n-1}\leq t\leq t_n`, where :math:`t_n= n \tau/N`.

    Each segment value :math:`w_n` is the average of samples of :math:`v(t)`
    at the midpoints of :math:`M` equally sized subsegments between
    :math:`t_{n-1}` and :math:`t_n`:

    .. math::
        w_n = \frac{1}{M}
        \sum_{m=1}^M v\left(t_{n-1} + \left(m-\tfrac{1}{2}\right) \frac{\tau}{MN} \right).

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create discretized Gaussian signal.

    >>> times = graph.identity_stf()
    >>> gaussian_signal = graph.exp(- (times - 5e-6) ** 2 / 2e-6 ** 2) / 2e-6
    >>> discretized_gamma_signal = graph.discretize_stf(
    ...     stf=gaussian_signal, duration=10e-6, segment_count=256, name="gamma"
    ... )
    >>> discretized_gamma_signal
    <Pwc: name="gamma", operation_name="discretize_stf", value_shape=(), batch_shape=()>

    Refer to the `How to create dephasing and amplitude robust single-qubit gates
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-create-dephasing-and-amplitude-robust-single-qubit-gates>`_
    user guide to find the example in context.
    """

    name = "discretize_stf"
    args = [
        forge.arg("stf", type=Stf),
        forge.arg("duration", type=float),
        forge.arg("segment_count", type=int),
        forge.arg("sample_count_per_segment", type=int, default=1),
    ]
    rtype = Pwc
    categories = [Category.FILTERING_AND_DISCRETIZING]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        stf = kwargs.get("stf")
        duration = kwargs.get("duration")
        check_duration(duration, "duration")
        segment_count = kwargs.get("segment_count")
        batch_shape, value_shape = validate_batch_and_value_shapes(stf, "stf")
        check_argument_integer(segment_count, "segment_count")
        check_argument(
            segment_count > 0,
            "The number of segments must be greater than zero.",
            {"segment_count": segment_count},
        )
        durations = duration / segment_count * np.ones(segment_count)
        sample_count_per_segment = kwargs.get("sample_count_per_segment")
        check_argument_integer(sample_count_per_segment, "sample_count_per_segment")
        check_argument(
            sample_count_per_segment > 0,
            "The number of samples per segment to take must be greater than zero.",
            {"sample_count_per_segment": sample_count_per_segment},
        )
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )


class TimeEvolutionOperatorsStf(Node):
    """
    Calculate the time-evolution operators for a system defined by an STF Hamiltonian by using a
    4th order Runge–Kutta method.

    Parameters
    ----------
    hamiltonian : Stf
        The control Hamiltonian, or batch of control Hamiltonians.
    sample_times : list or tuple or np.ndarray(1D, real)
        The N times at which you want to sample the unitaries. Must be ordered and contain
        at least one element. If you don't provide `evolution_times`, `sample_times` must
        start with 0.
    evolution_times : list or tuple or np.ndarray(1D, real) or None, optional
        The times at which the Hamiltonian should be sampled for the Runge–Kutta integration.
        If you provide it, must start with 0 and be ordered.
        If you don't provide it, the `sample_times` are used for the integration.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        Tensor of shape ``(..., N, D, D)``, representing the unitary time evolution.
        The n-th element (along the -3 dimension) represents the unitary (or batch of unitaries)
        from t = 0 to ``sample_times[n]``.

    See Also
    --------
    Graph.time_evolution_operators_pwc : Corresponding operation for `Pwc` Hamiltonians.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Simulate the dynamics of a qubit, where a simple Gaussian drive rotate the qubit
    along the x-axis.

    >>> duration = np.pi
    >>> initial_state = np.array([1, 0])
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> time = graph.identity_stf()
    >>> gaussian_drive = graph.exp(-(time ** 2))
    >>> hamiltonian = gaussian_drive * sigma_x * np.sqrt(np.pi) / 2
    >>> graph.time_evolution_operators_stf(
    ...     hamiltonian=hamiltonian,
    ...     sample_times=[duration],
    ...     evolution_times=np.linspace(0, duration, 200),
    ...     name="unitaries",
    ... )
    <Tensor: name="unitaries", operation_name="time_evolution_operators_stf", shape=(1, 2, 2)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="unitaries")
    >>> result["output"]["unitaries"]["value"].dot(initial_state)
    array([[0.70711169 + 0.0j, 0.0 - 0.70710187j]])
    """

    name = "time_evolution_operators_stf"
    args = [
        forge.arg("hamiltonian", type=Stf),
        forge.arg("sample_times", type=Union[list, tuple, np.ndarray]),
        forge.arg(
            "evolution_times",
            type=Optional[Union[list, tuple, np.ndarray]],
            default=None,
        ),
    ]
    rtype = Tensor
    categories = [Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        hamiltonian = kwargs.get("hamiltonian")
        sample_times = kwargs.get("sample_times")
        sample_times = np.asarray(sample_times)
        evolution_times = kwargs.get("evolution_times")
        check_argument(
            isinstance(hamiltonian, Stf),
            "The Hamiltonian must be an Stf.",
            {"hamiltonian": hamiltonian},
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(
            hamiltonian, "hamiltonian"
        )
        check_sample_times(sample_times, "sample_times")
        validate_hamiltonian(hamiltonian, "hamiltonian")
        time_count = len(sample_times)
        if evolution_times is not None:
            evolution_times = np.asarray(evolution_times)
            check_sample_times(evolution_times, "evolution_times")
            check_argument(
                evolution_times[0] == 0,
                "The first of the evolution times must be zero.",
                {"evolution_times": evolution_times},
            )
        else:
            check_argument(
                sample_times[0] == 0,
                "If you don't provide evolution times, the first of the sample"
                " times must be zero.",
                {"sample_times": sample_times},
            )
        shape = batch_shape + (time_count,) + value_shape
        return Tensor(_operation, shape=shape)


class ConvolvePwc(Node):
    r"""
    Create the convolution of a piecewise-constant function with a kernel.

    Parameters
    ----------
    pwc : Pwc
        The piecewise-constant function :math:`\alpha(t)` to convolve. You
        can provide a batch of functions, in which case the convolution is
        applied to each element of the batch.
    kernel : ConvolutionKernel
        The node representing the kernel :math:`K(t)`.

    Returns
    -------
    Stf
        The sampleable function representing the signal :math:`(\alpha * K)(t)`
        (or batch of signals, if you provide a batch of functions).

    See Also
    --------
    Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
    Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
    Graph.gaussian_convolution_kernel :
        Create a convolution kernel representing a normalized Gaussian.
    Graph.pwc : Create piecewise-constant functions.
    Graph.sample_stf : Sample an `Stf` at given times.
    Graph.sinc_convolution_kernel : Create a convolution kernel representing the sinc function.

    Notes
    -----
    The convolution is

    .. math::
        (\alpha * K)(t) \equiv
        \int_{-\infty}^\infty \alpha(\tau) K(t-\tau) d\tau.

    Convolution in the time domain is equivalent to multiplication in the
    frequency domain, so this function can be viewed as applying a linear
    time-invariant filter (specified via its time domain kernel :math:`K(t)`)
    to :math:`\alpha(t)`.

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Filter a piecewise-constant signal using a Gaussian convolution kernel.

    >>> gaussian_kernel = graph.gaussian_convolution_kernel(std=1.0, offset=3.0)
    >>> gaussian_kernel
    <ConvolutionKernel: operation_name="gaussian_convolution_kernel">
    >>> pwc_signal
    <Pwc: name="alpha", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    >>> filtered_signal = graph.convolve_pwc(pwc=pwc_signal, kernel=gaussian_kernel)
    >>> filtered_signal
    <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

    Refer to the `How to add smoothing and band-limits to optimized controls
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-add-smoothing-and-band-limits-to-optimized-controls>`_
    user guide to find the example in context.
    """

    name = "convolve_pwc"
    args = [forge.arg("pwc", type=Pwc), forge.arg("kernel", type=ConvolutionKernel)]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.FILTERING_AND_DISCRETIZING]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        pwc = kwargs.get("pwc")
        batch_shape, value_shape = validate_batch_and_value_shapes(pwc, "pwc")
        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)


class SincConvolutionKernel(Node):
    r"""
    Create a convolution kernel representing the sinc function.

    Use this kernel to eliminate angular frequencies above a certain cutoff.

    Parameters
    ----------
    cutoff_frequency : float or Tensor
        Upper limit :math:`\omega_c` of the range of angular frequencies that you want
        to preserve. The filter eliminates components of the signal that have
        higher angular frequencies.

    Returns
    -------
    ConvolutionKernel
        A node representing the sinc function to use in a convolution.

    See Also
    --------
    Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
    Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
    Graph.gaussian_convolution_kernel :
        Create a convolution kernel representing a normalized Gaussian.

    Notes
    -----
    The sinc kernel that this node represents is defined as

    .. math::
        K(t) = \frac{\sin(\omega_c t)}{\pi t}.

    In the frequency domain, the sinc function is constant in the range
    :math:`[-\omega_c, \omega_c]` and zero elsewhere. The filter it represents therefore
    passes angular frequencies only in that range.

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Filter a signal by convolving it with a sinc kernel.

    >>> sinc_kernel = graph.sinc_convolution_kernel(cutoff_frequency=300e6)
    >>> sinc_kernel
    <ConvolutionKernel: operation_name="sinc_convolution_kernel">
    >>> pwc_signal
    <Pwc: name="pwc_signal_#1", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    >>> filtered_signal = graph.convolve_pwc(pwc=pwc_signal, kernel=sinc_kernel)
    >>> filtered_signal
    <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

    Refer to the `How to create leakage-robust single-qubit gates
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-create-leakage-robust-single-qubit-gates>`_
    user guide to find the example in context.
    """

    name = "sinc_convolution_kernel"
    args = [forge.arg("cutoff_frequency", type=Union[float, Tensor])]
    kwargs = {}  # ConvolutionKernels don't accept name as an argument.
    rtype = ConvolutionKernel
    categories = [Category.FILTERING_AND_DISCRETIZING]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return ConvolutionKernel(_operation)


class GaussianConvolutionKernel(Node):
    r"""
    Create a convolution kernel representing a normalized Gaussian.

    Use this kernel to allow angular frequencies in the range roughly determined by
    its width, and progressively suppress components outside that range.

    Parameters
    ----------
    std : float or Tensor
        Standard deviation :math:`\sigma` of the Gaussian in the time domain.
        The standard deviation in the frequency domain is its inverse, so that
        a high value of this parameter lets fewer angular frequencies pass.
    offset : float or Tensor or None, optional
        Center :math:`\mu` of the Gaussian distribution in the time domain.
        Use this to offset the signal in time. Defaults to 0.

    Returns
    -------
    ConvolutionKernel
        A node representing a Gaussian function to use in a convolution.

    See Also
    --------
    Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
    Graph.sinc_convolution_kernel : Create a convolution kernel representing the sinc function.

    Notes
    -----
    The Gaussian kernel that this node represents is defined as:

    .. math::
        K(t) = \frac{e^{-(t-\mu)^2/(2\sigma^2)}}{\sqrt{2\pi\sigma^2}}.

    In the frequency domain, this Gaussian has standard deviation
    :math:`\omega_c= \sigma^{-1}`. The filter it represents therefore
    passes angular frequencies roughly in the range :math:`[-\omega_c, \omega_c]`.

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Filter a signal by convolving it with a Gaussian kernel.

    >>> gaussian_kernel = graph.gaussian_convolution_kernel(std=1.0, offset=3.0)
    >>> gaussian_kernel
    <ConvolutionKernel: operation_name="gaussian_convolution_kernel">
    >>> signal
    <Pwc: name="alpha", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    >>> filtered_signal = graph.convolve_pwc(pwc=signal, kernel=gaussian_kernel)
    >>> filtered_signal
    <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

    Refer to the `How to characterize a transmission line using a qubit as a probe
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-characterize-a-transmission-line-using-a-qubit-as-a-probe>`_
    user guide to find the example in context.
    """

    name = "gaussian_convolution_kernel"
    args = [
        forge.arg("std", type=Union[float, Tensor]),
        forge.arg("offset", type=Optional[Union[float, Tensor]], default=0),
    ]
    kwargs = {}  # ConvolutionKernels don't accept name as an argument.
    rtype = ConvolutionKernel
    categories = [Category.FILTERING_AND_DISCRETIZING]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return ConvolutionKernel(_operation)


class SampleStf(Node):
    """
    Sample an Stf at the given times.

    Parameters
    ----------
    stf : Stf
        The Stf to sample.
    sample_times : list or tuple or np.ndarray(1D, real)
        The times at which you want to sample the Stf. Must be ordered and contain
        at least one element.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The values of the Stf at the given times.

    See Also
    --------
    Graph.constant_stf_operator : Create a constant `Stf` operator.
    Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
    Graph.identity_stf : Create an `Stf` representing the identity function.
    Graph.sample_pwc : Sample a `Pwc` at given times.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    name = "sample_stf"
    args = [
        forge.arg("stf", type=Stf),
        forge.arg("sample_times", type=Union[list, tuple, np.ndarray]),
    ]
    rtype = Tensor
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        stf = kwargs.get("stf")
        check_argument(isinstance(stf, Stf), "The stf must be an Stf.", {"stf": stf})
        batch_shape, value_shape = validate_batch_and_value_shapes(stf, "stf")
        sample_times = kwargs.get("sample_times")
        sample_times = np.asarray(sample_times)
        check_sample_times(sample_times, "sample_times")
        time_count = len(sample_times)
        shape = batch_shape + (time_count,) + value_shape
        return Tensor(_operation, shape=shape)


class IdentityStf(Node):
    r"""
    Create an Stf representing the identity function, f(t) = t.

    Returns
    -------
    Stf
        An Stf representing the identity function.

    See Also
    --------
    Graph.constant_stf: Create a batch of constant `Stf`\s.
    Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
    Graph.sample_stf : Sample an `Stf` at given times.

    Notes
    -----
    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.

    Examples
    --------
    Create Gaussian pulse.

    >>> time = graph.identity_stf()
    >>> time
    <Stf: operation_name="identity_stf", value_shape=(), batch_shape=()>
    >>> gaussian = graph.exp(- time ** 2)
    >>> gaussian
    <Stf: operation_name="exp", value_shape=(), batch_shape=()>

    See more examples in the `How to define continuous analytical Hamiltonians
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-define-continuous-analytical-
    hamiltonians>`_ user guide.
    """

    name = "identity_stf"
    args = []
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return Stf(_operation, value_shape=(), batch_shape=())



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    Optional,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node

from . import node_data
from .documentation import Category
from .node_data import Pwc
from .utils import validate_filter_function_input


class FilterFunctionOperation(Node):
    r"""
    Evaluate the filter function for a control Hamiltonian and a noise operator
    at the given frequency elements.

    Parameters
    ----------
    control_hamiltonian : Pwc
        The control Hamiltonian :math:`H_\mathrm{c}(t)`.
    noise_operator : Pwc
        The noise operator :math:`N(t)`.
    frequencies : list or tuple or np.ndarray
        The elements in the frequency domain at which to return the values of the filter function.
    sample_count : int or None, optional
        The number of points in time, :math:`M`, to sample the control-frame noise operator.
        These samples are used to calculate the approximate Fourier integral efficiently.
        If None the piecewise Fourier integral is calculated exactly.
        Defaults to 100.
    projection_operator : np.ndarray or None, optional
        The projection operator :math:`P`. Defaults to the identity matrix.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    ~graphs.FilterFunction
        The filter function.

    See Also
    --------
    Graph.frequency_domain_noise_operator :
        Control-frame noise operator in the frequency domain.

    Notes
    -----
    The filter function is defined as [1]_:

    .. math::
        F(f) = \frac{1}{\mathrm{Tr}(P)} \mathrm{Tr} \left(
        P \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f)
        \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f)^\dagger P \right),

    with the control-frame noise operator in the frequency domain

    .. math::
        \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) = \int_0^\tau
        e^{-i 2\pi f t} \tilde N^\prime(t) \mathrm{d}t,

    where

    .. math::
        \tilde N^\prime(t) = \tilde N(t) - \frac{\mathrm{Tr}\left( P \tilde N(t)
        P \right)}{\mathrm{Tr}(P)} \mathbb{I}

    is the traceless control-frame noise operator in the time domain,

    .. math::
        \tilde N(t) = U_c^\dagger(t) N(t) U_c(t)

    is the control-frame noise operator in the time domain, and :math:`U_c(t)` is
    the time evolution induced by the control Hamiltonian.

    References
    ----------
    .. [1] `H. Ball, M. J. Biercuk, A. R. R. Carvalho, J. Chen, M. Hush,
            L. A. De Castro, L. Li, P. J. Liebermann, H. J. Slatyer, and C. Edmunds,
            Quantum Sci. Technol. 6, 044011 (2021).
            <https://doi.org/10.1088/2058-9565/abdca6>`_
    """

    name = "filter_function"
    args = [
        forge.arg("control_hamiltonian", type=Pwc),
        forge.arg("noise_operator", type=Pwc),
        forge.arg("frequencies", type=Union[list, tuple, np.ndarray]),
        forge.arg("sample_count", type=Optional[int], default=100),
        forge.arg("projection_operator", type=Optional[np.ndarray], default=None),
    ]
    rtype = node_data.FilterFunction
    categories = [Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        control_hamiltonian = kwargs.get("control_hamiltonian")
        noise_operator = kwargs.get("noise_operator")
        frequencies = kwargs.get("frequencies")
        frequencies = np.asarray(frequencies)
        sample_count = kwargs.get("sample_count")
        projection_operator = kwargs.get("projection_operator")
        _ = validate_filter_function_input(
            control_hamiltonian,
            noise_operator,
            frequencies,
            sample_count,
            projection_operator,
        )

        return node_data.FilterFunction(
            _operation, frequencies=frequencies, exact=sample_count is None
        )


class FrequencyDomainNoiseOperator(Node):
    r"""
    Create a control-frame noise operator in the frequency domain for a control Hamiltonian
    and a noise operator at the given frequencies.

    Parameters
    ----------
    control_hamiltonian : Pwc
        The control Hamiltonian :math:`H_\mathrm{c}(t)`.
    noise_operator : Pwc
        The noise operator :math:`N(t)`.
    frequencies : list or tuple or np.ndarray
        The elements in the frequency domain at which to return the values of the filter function.
    sample_count : int or None, optional
        The number of points in time, :math:`M`, to sample the control-frame noise operator.
        These samples are used to calculate the approximate Fourier integral efficiently.
        If None the piecewise Fourier integral is calculated exactly.
        Defaults to 100.
    projection_operator : np.ndarray or None, optional
        The projection operator :math:`P`. Defaults to the identity matrix.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The noise operator in the frequency domain.

    See Also
    --------
    Graph.filter_function :
        The filter function.

    Notes
    -----
    The control-frame noise operator in the frequency domain
    is defined as the Fourier transform of the operator in the time domain [1]_:

    .. math::
        \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) = \int_0^\tau
        e^{-i 2\pi f t} \tilde N^\prime(t) \mathrm{d}t,

    where

    .. math::
        \tilde N^\prime(t) = \tilde N(t) - \frac{\mathrm{Tr}\left( P \tilde N(t)
        P \right)}{\mathrm{Tr}(P)} \mathbb{I}

    is the traceless control-frame noise operator in the time domain,

    .. math::
        \tilde N(t) = U_c^\dagger(t) N(t) U_c(t)

    is the control-frame noise operator in the time domain, and :math:`U_c(t)` is
    the time evolution induced by the control Hamiltonian. If `sample_count` is
    set, the Fourier integral is approximated as

    .. math::
        \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) \approx \sum_{n=0}^{M-1}
        \frac{\tau}{M} e^{-i 2\pi f n \tau/M}
        \langle \tilde N^\prime (n\tau/M) \rangle,

    where :math:`\tau` is the duration of the control Hamiltonian.

    References
    ----------
    .. [1] `H. Ball, M. J. Biercuk, A. R. R. Carvalho, J. Chen, M. Hush,
            L. A. De Castro, L. Li, P. J. Liebermann, H. J. Slatyer, and C. Edmunds,
            Quantum Sci. Technol. 6, 044011 (2021).
            <https://doi.org/10.1088/2058-9565/abdca6>`_
    """

    name = "frequency_domain_noise_operator"
    args = [
        forge.arg("control_hamiltonian", type=Pwc),
        forge.arg("noise_operator", type=Pwc),
        forge.arg("frequencies", type=Union[list, tuple, np.ndarray]),
        forge.arg("sample_count", type=Optional[int], default=100),
        forge.arg("projection_operator", type=Optional[np.ndarray], default=None),
    ]
    rtype = node_data.Tensor
    categories = [Category.TIME_EVOLUTION]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        control_hamiltonian = kwargs.get("control_hamiltonian")
        noise_operator = kwargs.get("noise_operator")
        frequencies = kwargs.get("frequencies")
        frequencies = np.asarray(frequencies)
        sample_count = kwargs.get("sample_count")
        projection_operator = kwargs.get("projection_operator")
        dimension = validate_filter_function_input(
            control_hamiltonian,
            noise_operator,
            frequencies,
            sample_count,
            projection_operator,
        )
        count = len(frequencies)

        return node_data.Tensor(_operation, shape=(count, dimension, dimension))



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    Any,
    Optional,
    TypeVar,
)

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.node_data import Pwc
from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
)
from boulderopal.graph import (
    Graph,
    execute_graph,
)
from boulderopal.ions._drives import (
    Drive,
    OptimizableDrive,
)
from boulderopal.optimization import run_optimization


def _validate_system_parameters(
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: Optional[np.ndarray],
) -> int:
    """
    Validate the arrays describing an ion system
    and return the number of ions.
    """

    ion_count = lamb_dicke_parameters.shape[-1]

    check_argument(
        lamb_dicke_parameters.shape == (3, ion_count, ion_count)
        and relative_detunings.shape == (3, ion_count),
        "The shape of the Lamb–Dicke parameters array must be (3, N, N), "
        "and the shape of the relative detunings array must be (3, N), "
        "where N is the number of ions.",
        {
            "lamb_dicke_parameters": lamb_dicke_parameters,
            "relative_detunings": relative_detunings,
        },
        extras={
            "lamb_dicke_parameters.shape": lamb_dicke_parameters.shape,
            "relative_detunings.shape": relative_detunings.shape,
        },
    )

    if target_phases is not None:
        check_argument(
            target_phases.shape == (ion_count, ion_count),
            "The shape of the target phases array must be (N, N), "
            "where N is the number of ions.",
            {"target_phases": target_phases},
            extras={"ion count": ion_count, "target_phases.shape": target_phases.shape},
        )

    return ion_count


_T = TypeVar("_T", Drive, OptimizableDrive)


def _check_drives_addressing(drives: list[_T], ion_count: int) -> None:
    """
    Check the input drives are a list and that the ions they address are valid.
    """

    check_argument(
        isinstance(drives, list),
        "You must provide a list of drives.",
        {"drives": drives},
    )

    all_addressing: list[int] = []
    for idx, drive in enumerate(drives):
        check_argument(
            all(0 <= ion < ion_count for ion in drive.addressing),
            "The addressed ions must be between 0 (inclusive) "
            "and the number of ions (exclusive).",
            {"drives": drives},
            extras={
                f"drives[{idx}].addressing": drive.addressing,
                "ion count": ion_count,
            },
        )
        all_addressing.extend(drive.addressing)

    check_argument(
        len(all_addressing) == len(set(all_addressing)),
        "Each ion can only be addressed by a single drive.",
        {"drives": drives},
    )


def _get_ion_drives(
    pwc_addressing_pairs: list[tuple[Pwc, tuple[int, ...]]],
    ion_count: int,
    graph: Graph,
    duration: float,
) -> list[Pwc]:
    """
    From a list of (Pwc, list(int)) tuples (drives and ions addressed by them),
    return a list of length ion_count the drive addressing each ion
    or a PWC with value 0 if the ion is not addressed by any drive.
    """
    ion_drives = []
    for idx in range(ion_count):
        for pwc, addressing in pwc_addressing_pairs:
            # Add the first drive that addresses the ion as we assume
            # each ion can only be addressed by a single drive.
            if idx in addressing:
                ion_drives.append(pwc)
                break
        else:
            ion_drives.append(graph.constant_pwc(constant=0.0, duration=duration))

    return ion_drives


_MS_NODE_NAMES = ["sample_times", "phases", "displacements", "infidelities"]


def ms_simulate(
    drives: list[Drive],
    duration: float,
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: Optional[np.ndarray] = None,
    sample_count: int = 128,
) -> dict:
    r"""
    Simulate a Mølmer–Sørensen-type operation on a system composed of ions.

    This function builds a graph describing the Mølmer–Sørensen operation
    and calls boulderopal.execute_graph to simulate the ion dynamics.

    Parameters
    ----------
    drives : list[~ions.Drive]
        A list of drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    duration : float
        The duration, in seconds, of the dynamics to be simulated, :math:`T`.
        It must be greater than zero.
    lamb_dicke_parameters : np.ndarray
        A 3D array of shape ``(3, N, N)``, where :math:`N` is the number of
        ions in the system, specifying the laser-ion coupling strength,
        :math:`\{\eta_{jkl}\}`. The three dimensions indicate, respectively,
        the axis, the collective mode number, and the ion.
    relative_detunings : np.ndarray
        A 2D array of shape ``(3, N)`` specifying the difference, in Hz, between
        each motional mode frequency and the laser detuning
        (with respect to the qubit transition frequency :math:`\omega_0`),
        :math:`\{\delta_{jk} = \nu_{jk} - \delta\}`. The two dimensions indicate,
        respectively, the axis and the collective mode number.
    target_phases : np.ndarray or None, optional
        A 2D array of shape ``(N, N)`` with the target relative phases between
        ion pairs, :math:`\{\Psi_{kl}\}`, as a strictly lower triangular matrix.
        Its :math:`(k, l)`-th element indicates the total relative phase target
        for ions :math:`k` and :math:`l`, with :math:`k > l`.
        If not provided, the function does not return the operational infidelities.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (included)
        at which the evolution is sampled.
        Defaults to 128.

    Returns
    -------
    dict
        The result of the `execute_graph` call.
        Its ``output`` item is a dictionary containing information about
        the evolution of the system, with the following keys:

            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{jk}(t_i) = \phi_{jk}(t_i) + \phi_{kj}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{jk}` records the relative phase between ions :math:`j`
                and :math:`k`; matrix elements where :math:`j \leq k` are zero.
            ``displacements``
                Displacements :math:`\{\eta_{pj}\alpha_{pj}(t_i)\}` for all mode-ion
                combinations, as an array of shape ``(T, 3, N, N)``.
                The four dimensions indicate the sample time, the axis,
                the collective mode number, and the ion.
            ``infidelities``
                A 1D array of length ``T`` representing the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.
                Only returned if target relative phases are provided.

    See Also
    --------
    boulderopal.ions.Drive :
        Class describing non-optimizable drives.
    boulderopal.ions.ms_optimize :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    boulderopal.ions.obtain_ion_chain_properties :
        Calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`,
    where :math:`\{\gamma_j\}` is the total drive acting on ion :math:`j`.

    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
            - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    The operational infidelity of the Mølmer–Sørensen gate is defined as [1]_:

    .. math::
        \mathcal{I} = 1 - \left| \left( \prod_{\substack{k=1 \\ l<k}}^N \cos (
            \phi_{kl} - \psi_{kl}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k,l=1}^N \left[ |\eta_{jkl}|^2
            |\alpha_{jkl}|^2 \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    duration = ScalarDType.REAL(duration, "duration", min_=0)

    lamb_dicke_parameters = ArrayDType.REAL(
        lamb_dicke_parameters, "lamb_dicke_parameters", ndim=3
    )
    relative_detunings = ArrayDType.REAL(
        relative_detunings, "relative_detunings", ndim=2
    )
    target_phases = nullable(ArrayDType.REAL, target_phases, "target_phases", ndim=2)
    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    check_argument(
        all(isinstance(drive, Drive) for drive in drives),
        "All drives must be non-optimizable.",
        {"drives": drives},
    )

    _check_drives_addressing(drives, ion_count)

    graph = Graph()

    drive_pwcs = [
        (drive.get_pwc(graph, duration), drive.addressing) for drive in drives
    ]
    ion_drives = _get_ion_drives(drive_pwcs, ion_count, graph, duration)

    sample_times = np.linspace(0.0, duration, sample_count)
    graph.tensor(sample_times, name="sample_times")

    phases = graph.ions.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    displacements = graph.ions.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[2],
    )

    if target_phases is not None:
        graph.ions.ms_infidelity(
            phases=phases,
            displacements=displacements,
            target_phases=target_phases,
            name=_MS_NODE_NAMES[3],
        )
        output_node_names = _MS_NODE_NAMES
    else:
        output_node_names = _MS_NODE_NAMES[:3]

    return execute_graph(graph=graph, output_node_names=output_node_names)


def ms_optimize(
    drives: list[OptimizableDrive],
    duration: float,
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: np.ndarray,
    sample_count: int = 128,
    robust: bool = False,
    **optimization_kwargs: Any,
) -> dict:
    r"""
    Find optimal pulses to perform a target Mølmer–Sørensen-type operation
    on a system composed of ions.

    This function builds a graph describing the Mølmer–Sørensen operation
    and calls boulderopal.run_optimization to minimize the target cost.

    Parameters
    ----------
    drives : list[OptimizableDrive]
        A list of optimizable drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    duration : float
        The duration, in seconds, of the dynamics to be optimized, :math:`T`.
        It must be greater than zero.
    lamb_dicke_parameters : np.ndarray
        A 3D array of shape ``(3, N, N)``, where :math:`N` is the number of
        ions in the system, specifying the laser-ion coupling strength,
        :math:`\{\eta_{jkl}\}`. The three dimensions indicate, respectively,
        the axis, the collective mode number, and the ion.
    relative_detunings : np.ndarray
        A 2D array of shape ``(3, N)`` specifying the difference, in Hz, between
        each motional mode frequency and the laser detuning
        (with respect to the qubit transition frequency :math:`\omega_0`),
        :math:`\{\delta_{jk} = \nu_{jk} - \delta\}`. The two dimensions indicate,
        respectively, the axis and the collective mode number.
    target_phases : np.ndarray
        A 2D array of shape ``(N, N)`` with the target relative phases between
        ion pairs, :math:`\{\Psi_{kl}\}`, as a strictly lower triangular matrix.
        Its :math:`(k, l)`-th element indicates the total relative phase target
        for ions :math:`k` and :math:`l`, with :math:`k > l`.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (both included)
        at which the evolution is sampled.
        Defaults to 128.
    robust : bool, optional
        If set to False, the cost corresponds to the infidelity at the end of the gate.
        If set to True, the cost is the final infidelity plus a dephasing-robust cost term.
        Defaults to False.
    **optimization_kwargs
        Additional parameters to pass to boulderopal.run_optimization.

    Returns
    -------
    dict
        The result of the `run_optimization` call.
        Its ``output`` item is a dictionary containing information about
        the optimized drive and the evolution of the system, with the following keys:

            optimized drives
                The piecewise-constant optimized drives implementing the gate.
                The keys are the names of the `drives` provided to the function.
            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{jk}(t_i) = \phi_{jk}(t_i) + \phi_{kj}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{jk}` records the relative phase between ions :math:`j`
                and :math:`k`; matrix elements where :math:`j \leq k` are zero.
            ``displacements``
                Displacements :math:`\{\eta_{pj}\alpha_{pj}(t_i)\}` for all mode-ion
                combinations, as an array of shape ``(T, 3, N, N)``.
                The four dimensions indicate the sample time, the axis,
                the collective mode number, and the ion.
            ``infidelities``
                A 1D array of length ``T`` representing the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.

    See Also
    --------
    boulderopal.ions.ComplexOptimizableDrive :
        Class describing a piecewise-constant complex-valued optimizable drive.
    boulderopal.ions.RealOptimizableDrive :
        Class describing a piecewise-constant real-valued optimizable drive.
    boulderopal.ions.ms_simulate :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.
    boulderopal.ions.obtain_ion_chain_properties :
        Calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`,
    where :math:`\{\gamma_j\}` is the total drive acting on ion :math:`j`.

    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
            - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    The operational infidelity of the Mølmer–Sørensen gate is defined as [1]_:

    .. math::
        \mathcal{I} = 1 - \left| \left( \prod_{\substack{k=1 \\ l<k}}^N \cos (
            \phi_{kl} - \psi_{kl}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k,l=1}^N \left[ |\eta_{jkl}|^2
            |\alpha_{jkl}|^2 \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 .

    You can use the `robust` flag to construct a Mølmer–Sørensen gate that is
    robust against dephasing noise. This imposes a symmetry [1]_ in the optimizable
    ion drives and aims to minimize the time-averaged positions of the phase-space
    trajectories,

    .. math::
        \langle \alpha_{pj} \rangle
            = \frac{1}{t_\text{gate}} \int_0^{t_\text{gate}}
                \alpha_{pj}(t) \mathrm{d} t ,

    where the axis dimension and the collective mode dimension are combined
    into a single index :math:`p` for simplicity.

    This is achieved by adding an additional term to the cost function,
    consisting of the sum of the square moduli of the time-averaged positions
    multiplied by the corresponding Lamb–Dicke parameters. That is to say,

    .. math::
        C_\text{robust} =
            \mathcal{I} + \sum_{p,j}
                \left| \eta_{pj} \langle \alpha_{pj} \rangle \right|^2 .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    duration = ScalarDType.REAL(duration, "duration", min_=0)

    lamb_dicke_parameters = ArrayDType.REAL(
        lamb_dicke_parameters, "lamb_dicke_parameters", ndim=3
    )
    relative_detunings = ArrayDType.REAL(
        relative_detunings, "relative_detunings", ndim=2
    )
    target_phases = ArrayDType.REAL(target_phases, "target_phases", ndim=2)
    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    check_argument(
        all(isinstance(drive, OptimizableDrive) for drive in drives),
        "All drives must be optimizable.",
        {"drives": drives},
    )

    _check_drives_addressing(drives, ion_count)

    drive_names = [drive.name for drive in drives]

    check_argument(
        len(drive_names) == len(set(drive_names)),
        "The drive names must be unique.",
        {"drives": drives},
        extras={"[drive.name for drive in drives]": drive_names},
    )

    graph = Graph()

    drive_pwcs = [
        (drive.get_pwc(graph, duration, robust), drive.addressing) for drive in drives
    ]
    ion_drives = _get_ion_drives(drive_pwcs, ion_count, graph, duration)

    sample_times = np.linspace(0.0, duration, sample_count)
    graph.tensor(sample_times, name="sample_times")

    phases = graph.ions.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    displacements = graph.ions.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[2],
    )

    infidelities = graph.ions.ms_infidelity(
        phases=phases,
        displacements=displacements,
        target_phases=target_phases,
        name=_MS_NODE_NAMES[3],
    )

    cost = infidelities[-1]
    if robust:
        cost += graph.ions.ms_dephasing_robust_cost(
            drives=ion_drives,
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )

    return run_optimization(
        graph=graph,
        cost_node_name=cost.name,
        output_node_names=drive_names + _MS_NODE_NAMES,
        **optimization_kwargs,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from datetime import datetime
from typing import (
    Any,
    Optional,
)

from rich.console import Console
from rich.table import Table

from boulderopal._configuration.cloud import get_cloud_router
from boulderopal._validation.basic import ScalarDType

_ACTION_NAME_MAP = {
    "closed_loop_step_workflow": "closed_loop.step",
    "execute_graph_workflow": "execute_graph",
    "obtain_ion_chain_properties_workflow": "ions.obtain_ion_chain_properties",
    "reconstruct_noise_workflow": "noise_reconstruction.reconstruct",
    "run_gradient_free_optimization_workflow": "run_gradient_free_optimization",
    "run_optimization_workflow": "run_optimization",
    "run_stochastic_optimization_workflow": "run_stochastic_optimization",
}


def show_activity_monitor(
    limit: int = 5, offset: int = 0, status: Optional[str] = None
) -> None:
    """
    Print information from previously submitted actions.

    Parameters
    ----------
    limit : int, optional
        The number of actions to print. Cannot exceed 50. Defaults to 5.
    offset : int, optional
        The number of recent actions to ignore before starting to print.
        Defaults to 0.
    status : str or None, optional
        The filter for action status. Defaults to None, meaning actions with all status are
        printed. Available options are SUCCESS, FAILURE, REVOKED, PENDING,
        RECEIVED, RETRY, and STARTED.
    """

    limit = ScalarDType.INT(limit, name="limit", max_=50, max_inclusive=True)
    router = get_cloud_router("Activity monitor")
    action_records: list[dict[str, Any]] = router.activity_monitor(
        limit, offset, status
    )

    if len(action_records) == 0:
        print("No actions found for the specified arguments.")
        return

    table = Table(title="Activity monitor", show_lines=True)
    console = Console()

    for header in ["Action name", "Created UTC", "Updated UTC", "Run time", "Status"]:
        table.add_column(header, justify="center")

    for row in action_records:
        table.add_row(*_format_action_record(row))

    console.print(table)


def _format_action_record(record: dict[str, str]) -> list[str]:
    created_at = datetime.fromisoformat(record["createdAt"])
    updated_at = datetime.fromisoformat(record["updatedAt"])
    run_time = int((updated_at - created_at).total_seconds())

    _state_map = {
        "SUCCESS": "Completed",
        "FAILURE": "Failed",
        "REVOKED": "Cancelled",
        "PENDING": "Queued",
        "RECEIVED": "Received",
        "RETRY": "Retried",
        "STARTED": "Started",
    }

    _name = _ACTION_NAME_MAP.get(record["name"], record["name"])

    return [
        f'{_name}\nID {record["modelId"]}',
        created_at.strftime("%Y-%m-%d\n%H:%M:%S"),
        updated_at.strftime("%Y-%m-%d\n%H:%M:%S"),
        f"{run_time//3600:02d}:{(run_time % 3600)//60:02d}:{run_time % 60:02d}",
        _state_map.get(record["status"], record["status"]),
    ]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


from __future__ import annotations

from enum import Enum
from typing import Optional

from boulderopal._core import run_optimization_workflow
from boulderopal._validation.basic import (
    ScalarDType,
    nullable,
    validate_enum,
)
from boulderopal._validation.graph import (
    check_cost_node,
    check_cost_node_for_optimization_graph,
    check_initial_value_for_optimization_node,
    check_optimization_node_in_graph,
    validate_output_node_names,
)
from boulderopal.graph._graph import Graph


class HistoryScope(Enum):
    """
    Configuration for the history data returned from a graph-based optimizer.

    Attributes
    ----------
    NONE
        Do not return any history information.

    ITERATION_VALUES
        Return the value obtained at each iteration and optimization.

    HISTORICAL_BEST
        Return the best values reached up to each iteration for each optimization.

    ALL
        Return both the values obtained at each iteration and the historical best values.

    See also
    --------
    boulderopal.run_gradient_free_optimization :
        Perform model-based optimization without using gradient values.
    boulderopal.run_optimization :
        Perform gradient-based deterministic optimization of generic real-valued functions.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.
    """

    NONE = "NONE"
    ITERATION_VALUES = "ITERATION_VALUES"
    HISTORICAL_BEST = "HISTORICAL_BEST"
    ALL = "ALL"


def run_optimization(
    graph: Graph,
    cost_node_name: str,
    output_node_names: str | list[str],
    max_iteration_count: Optional[int] = None,
    target_cost: Optional[float] = None,
    cost_tolerance: Optional[float] = None,
    optimization_count: int = 4,
    cost_history_scope: HistoryScope = HistoryScope.NONE,
    seed: Optional[int] = None,
) -> dict:
    r"""
    Perform gradient-based deterministic optimization of generic real-valued functions.

    Use this function to determine a choice of variables that minimizes the value of a scalar
    real-valued cost function of those variables. You express that cost function as a graph
    describing how the input variables are transformed into the cost value.

    Note that this function will pick a different initial guess for each optimization run.
    You can provide initial values (or a list of them of the same length) for the optimization
    variables in the graph, and the optimizer will start with those values.
    If the `optimization_count` is larger than the number of provided initial values,
    the optimizer will use random values after consuming all available initial values.

    Parameters
    ----------
    graph : Graph
        The graph describing the cost :math:`C(\mathbf v)` and outputs :math:`\{F_j(\mathbf v)\}`
        as functions of the optimizable input variables :math:`\mathbf v`.
        The graph must contain nodes with names :math:`s` (giving the cost function)
        and :math:`\{s_j\}` (giving the output functions).
    cost_node_name : str
        The name :math:`s` of the real-valued scalar graph node that defines the cost
        function :math:`C(\mathbf v)` to be minimized.
    output_node_names : str or list[str]
        The names :math:`\{s_j\}` of the graph nodes that define the output functions
        :math:`\{F_j(\mathbf v)\}`. The function evaluates these using the optimized
        variables and returns them in the output.
    max_iteration_count : int or None, optional
        The maximum number of cost evaluations to perform. You can set this as an early stop
        condition for the optimizer. Defaults to None, which means that the optimizer runs
        until it converges.
    target_cost : float or None, optional
        A target value of the cost that you can set as an early stop condition for the optimizer.
        If the cost becomes equal or smaller than this value, the optimization halts.
        Defaults to None, meaning that the optimizer runs until it converges.
    cost_tolerance : float or None, optional
        The tolerance :math:`\mathrm{tol}` to check whether the cost function has converged.
        You can set this as an early stop condition for the optimizer.
        The optimizer stops when the improvement of the cost function over the iterations is below
        the tolerance. That is, it stops when
        :math:`(C_j - C_{j+1}) / \max(|C_j|, |C_{j+1}|, 1) \le \mathrm{tol}`.
        Defaults to None, meaning the optimizer handles the tolerance automatically.
        You can try different values depending on your optimization problem.
        A recommended starting value is about 1e-9.
    optimization_count : int, optional
        The number :math:`N` of independent randomly seeded optimizations to perform.
        The function returns the results from the best optimization (the one with
        the lowest cost). Defaults to 4. Depending on the landscape of the optimization problem,
        a larger value will help in finding lower costs, at the expense of prolonging computation
        time.
    cost_history_scope : HistoryScope, optional
        Configuration for the scope of the returned cost history data.
        Use this to select if you want the history data to be returned.
        Defaults to no cost history data returned.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.
        Note that if your graph contains nodes generating random samples,
        you need to also set seeds for those nodes to ensure a reproducible result.

    Returns
    -------
    dict
        A dictionary containing the optimization result, with the following keys:

        ``cost``
            The minimum cost function value :math:`C(\mathbf v_\mathrm{optimized})`
            achieved across all optimizations.
        ``output``
            The dictionary giving the value of each requested output node, evaluated
            at the optimized variables, namely
            :math:`\{s_j: F_j(\mathbf v_\mathrm{optimized})\}`.
            The keys of the dictionary are the names :math:`\{s_j\}` of the output nodes.
        ``cost_history``
            The evolution of the cost function across all optimizations and iterations.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    boulderopal.closed_loop.optimize :
        Run a closed-loop optimization to find a minimum of the given cost function.
    boulderopal.closed_loop.step :
        Perform a single step in a closed-loop optimization.
    boulderopal.execute_graph :
        Evaluate generic functions.
    boulderopal.run_gradient_free_optimization :
        Perform model-based optimization without using gradient values.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.

    Notes
    -----
    Given a cost function :math:`C(\mathbf v)` of variables :math:`\mathbf v`, this function
    computes an estimate :math:`\mathbf v_\mathrm{optimized}` of
    :math:`\mathrm{argmin}_{\mathbf v} C(\mathbf v)`, namely the choice of variables
    :math:`\mathbf v` that minimizes :math:`C(\mathbf v)`. The function then calculates the values
    of arbitrary output functions :math:`\{F_j(\mathbf v_\mathrm{optimized})\}` with that choice
    of variables.

    This function represents the cost and output functions as nodes of a graph.
    This graph defines the input optimization variables :math:`\mathbf v`, and how these variables
    are transformed into the corresponding cost and output quantities. You build the graph from
    primitive nodes defined in the graph of the Boulder Opal package.
    Each such node, which can be identified by a name, represents a function of the previous nodes
    in the graph (and thus, transitively, a function of the input variables).
    You can use any named scalar real-valued node :math:`s` as the cost function, and any named
    nodes :math:`\{s_j\}` as outputs.

    After you provide a cost function :math:`C(\mathbf v)` (via a graph), this function runs an
    optimization process :math:`N` times, each with random initial variables, to identify :math:`N`
    local minima of the cost function, and then takes the variables corresponding to the best
    such minimum as :math:`\mathbf v_\mathrm{optimized}`.

    A common use-case for this function is to determine controls for a quantum system that yield
    an optimal gate: the variables :math:`\mathbf v` parameterize the controls to be optimized,
    and the cost function :math:`C(\mathbf v)` is the operational infidelity describing the quality
    of the resulting gate relative to a target gate. When combined with the node definitions in the
    Boulder Opal package, which make it convenient to define such cost functions, this function
    provides a highly configurable framework for quantum control that encapsulates other common
    tools such as gradient ascent pulse engineering [1]_ and chopped random basis (CRAB)
    optimization [2]_.

    The optimizer will carry out the gradient-based optimization until it converges to a minimum.
    You can set up early stop conditions such as a target cost or a maximum number of cost
    evaluations via the `target_cost` and `max_iteration_count` parameters. If both are provided,
    the optimizer will terminate when either of the two conditions is satisfied.

    References
    ----------
    .. [1] `N. Khaneja, T. Reiss, C. Kehlet, T. Schulte-Herbrüggen, and S. J. Glaser,
            Journal of Magnetic Resonance 172, 2 (2005).
            <https://doi.org/10.1016/j.jmr.2004.11.004>`_

    .. [2] `P. Doria, T. Calarco, and S. Montangero,
            Physical Review Letters 106, 190501 (2011).
            <https://doi.org/10.1103/PhysRevLett.106.190501>`_

    Examples
    --------
    Perform a simple optimization with variables that are initialized to given values.

    >>> graph = bo.Graph()
    >>> variables = graph.optimization_variable(
    ...     2, -1, 1, initial_values=np.array([0.6, 0.]), name="variables"
    ... )
    >>> x, y = variables[0], variables[1]
    >>> cost = (x - 0.5) ** 2 + (y - 0.1) ** 2
    >>> cost.name = "cost"
    >>> result = bo.run_optimization(
    ...     graph=graph, cost_node_name="cost", output_node_names="variables"
    ... )
    >>> result["cost"], result["output"]
        (0.0, {'variables': {'value': array([0.5, 0.1])}})

    See also the `How to optimize controls in arbitrary quantum systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-
    optimize-controls-in-arbitrary-quantum-systems-using-graphs>`_ user guide.
    """

    check_optimization_node_in_graph(graph)
    check_initial_value_for_optimization_node(graph)
    check_cost_node(cost_node_name, graph)

    output_node_names = validate_output_node_names(output_node_names, graph)
    check_cost_node_for_optimization_graph(graph, cost_node_name, output_node_names)

    result = run_optimization_workflow(
        graph=graph,
        optimization_count=ScalarDType.INT(
            optimization_count, "optimization_count", min_=0
        ),
        cost_node_name=cost_node_name,
        output_node_names=output_node_names,
        target_cost=nullable(ScalarDType.REAL, target_cost, "target_cost"),
        max_iteration_count=nullable(
            ScalarDType.INT, max_iteration_count, "max_iteration_count", min_=0
        ),
        cost_tolerance=nullable(
            ScalarDType.REAL, cost_tolerance, "cost_tolerance", min_=0
        ),
        cost_history_scope=validate_enum(HistoryScope, cost_history_scope),
        seed=nullable(ScalarDType.INT, seed, "seed", min_=0, min_inclusive=True),
    )
    return result



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from functools import wraps
from typing import (
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_argument_positive_scalar,
    check_argument_real_vector,
)

from .documentation import Category
from .namespace import create_namespace_doc
from .node_data import Sequence as Sequence_
from .node_data import (
    Stf,
    Tensor,
)
from .types import TensorLike


class RandomColoredNoiseStfSignal(Node):
    r"""
    Sample the one-sided power spectral density (PSD) of a random noise process in the
    time domain and returns the resultant noise trajectories as an Stf.

    Parameters
    ----------
    power_spectral_density : np.ndarray or Tensor (1D, real)
        The one-sided power spectral density of the noise sampled at frequencies
        :math:`\{0, \Delta f, 2\Delta f, \ldots , M\Delta f\}`.
    frequency_step : float
        The step size :math:`\Delta f` of power spectrum densities samples
       `power_spectral_density`. Must be a strictly positive number.
    batch_shape : list[int] or tuple[int], optional
        The batch shape of the returned Stf. By default, the batch shape is ``()``, that is,
        the returned Stf represents only one noise trajectory. If the batch shape is
        ``(m, n,...)``, the returned Stf represents `m*n*...` trajectories arranged in
        this batch shape.
    seed : int or None, optional
        A seed for the random number generator used for sampling. When set, same
        trajectories are produced on every run of this function, provided all the other
        arguments also remain unchanged. Defaults to None, in which case the
        generated noise trajectories can be different from one run to another.

    Returns
    -------
    Stf
        An `Stf` signal representing the noise trajectories in the time domain. The
        batch shape of this `Stf` is same as the argument `batch_shape`.

    Notes
    -----
    Given a frequency step size of :math:`\Delta f` and discrete samples
    :math:`P[k] = P(k\Delta f)` of a one-sided power spectral density function
    :math:`P(f)`, the output is a possibly batched Stf which represents one random
    realization of the random noise process. Each such trajectory is periodic with a
    time period of :math:`1/\Delta f`.

    Examples
    --------
    Create a PWC signal by sampling from a power spectral density function
    to define a noise Hamiltonian.

    >>> sigma_z = np.diag([1, -1])
    >>> frequency_step = 2e3
    >>> frequencies = np.arange(0, 2e6, frequency_step)
    >>> frequency_cutoff = 0.05e6
    >>> power_densities = 4e9 / (frequencies + frequency_cutoff)
    >>> noise_stf = graph.random.colored_noise_stf_signal(
    ...     power_spectral_density=power_densities, frequency_step=2000.0, batch_shape=(100,)
    ... )
    >>> noise_stf
    <Stf: operation_name="random_colored_noise_stf_signal", value_shape=(), batch_shape=(100,)>
    >>> noise_pwc = graph.discretize_stf(stf=noise_stf, duration=2e-6, segment_count=50)
    >>> noise_hamiltonian = noise_pwc * sigma_z
    >>> noise_hamiltonian
    <Pwc: name="multiply_#3", operation_name="multiply", value_shape=(2, 2), batch_shape=(100,)>

    Refer to the `How to simulate quantum dynamics subject to noise with graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-quantum-dynamics-
    subject-to-noise-with-graphs>`_ user guide to find the example in context.
    """

    name = "random_colored_noise_stf_signal"

    args = [
        forge.arg("power_spectral_density", type=TensorLike),
        forge.arg("frequency_step", type=float),
        forge.arg("batch_shape", type=Union[List[int], Tuple[int, ...]], default=()),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = Stf
    categories = [Category.RANDOM_OPERATIONS, Category.BUILDING_SMOOTH_HAMILTONIANS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        power_spectral_density = kwargs.get("power_spectral_density")
        frequency_step = kwargs.get("frequency_step")
        batch_shape = kwargs.get("batch_shape")
        seed = kwargs.get("seed")

        check_argument_real_vector(power_spectral_density, "power_spectral_density")
        check_argument_positive_scalar(frequency_step, "frequency_step")
        check_argument(
            isinstance(batch_shape, (list, tuple))
            and all(
                isinstance(x, (int, np.int32, np.int64)) and x > 0 for x in batch_shape
            ),
            "batch_shape should be a list or a tuple of positive integers.",
            {"batch_shape": batch_shape},
        )
        if seed is not None:
            check_argument_integer(seed, "seed")

        return Stf(_operation, value_shape=(), batch_shape=batch_shape)


class RandomNormal(Node):
    r"""
    Create a sample of normally distributed random numbers.

    Parameters
    ----------
    shape : tuple or list
        The shape of the sampled random numbers.
    mean : int or float
        The mean of the normal distribution.
    standard_deviation : int or float
        The standard deviation of the normal distribution.
    seed : int or None, optional
        A seed for the random number generator. Defaults to None,
        in which case a random value for the seed is used.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        A tensor containing a sample of normally distributed random numbers
        with shape ``shape``.

    See Also
    --------
    :func:`Graph.random.choices <random.choices>`
        Create random samples from the data that you provide.
    :func:`Graph.random.uniform <random.uniform>`
        Create a sample of uniformly distributed random numbers.
    :func:`boulderopal.run_stochastic_optimization`
        Function to find the minimum of generic stochastic functions.

    Examples
    --------
    Create a random tensor by sampling from a Gaussian distribution.

    >>> samples = graph.random.normal(
    ...     shape=(3, 1), mean=0.0, standard_deviation=0.05, seed=0, name="samples"
    ... )
    >>> result = bo.execute_graph(graph=graph, output_node_names="samples")
    >>> result["output"]["samples"]["value"]
    array([[-0.03171833], [0.00816805], [-0.06874011]])

    Create a batch of noise signals to construct a PWC Hamiltonian. The signal is defined
    as :math:`a \cos(\omega t)`, where :math:`a` follows a normal distribution and :math:`\omega`
    follows a uniform distribution.

    >>> seed = 0
    >>> batch_size = 3
    >>> sigma_x = np.array([[0, 1], [1, 0]])
    >>> sample_times = np.array([0.1, 0.2])
    >>> a = graph.random.normal((batch_size, 1), mean=0.0, standard_deviation=0.05, seed=seed)
    >>> omega = graph.random.uniform(
    ...     shape=(batch_size, 1), lower_bound=np.pi, upper_bound=2 * np.pi, seed=seed
    ... )
    >>> sampled_signal = a * graph.cos(omega * sample_times[None])
    >>> hamiltonian = graph.pwc_signal(sampled_signal, duration=0.2) * sigma_x
    >>> hamiltonian.name = "hamiltonian"
    >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
    >>> result["output"]["hamiltonian"]
    {
        'durations': array([0.1, 0.1]),
        'values': array([
            [
                [[-0.        , -0.02674376], [-0.02674376, -0.        ]],
                [[-0.        , -0.01338043], [-0.01338043, -0.        ]]
            ],
            [
                [[ 0.        ,  0.00691007], [ 0.00691007,  0.        ]],
                [[ 0.        ,  0.00352363], [ 0.00352363,  0.        ]]],
            [
                [[-0.        , -0.06230612], [-0.06230612, -0.        ]],
                [[-0.        , -0.04420857], [-0.04420857, -0.        ]]
            ]
        ]),
        'time_dimension': 1
    }

    See more examples in the `How to optimize controls robust to strong noise sources
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-robust-
    to-strong-noise-sources>`_ user guide.
    """

    name = "random_normal"

    args = [
        forge.arg("shape", type=Union[Tuple, List]),
        forge.arg("mean", type=Union[int, float]),
        forge.arg("standard_deviation", type=Union[int, float]),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    rtype = Tensor
    categories = [Category.RANDOM_OPERATIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        shape = kwargs.get("shape")
        standard_deviation = kwargs.get("standard_deviation")
        seed = kwargs.get("seed")

        check_argument(
            isinstance(shape, (tuple, list)),
            "The type of `shape` must be tuple or list.",
            {"shape": shape},
        )

        check_argument_positive_scalar(standard_deviation, "standard_deviation")

        if seed is not None:
            check_argument_integer(seed, "seed")

        return Tensor(_operation, shape=tuple(shape))


class RandomUniform(Node):
    r"""
    Create a sample of uniformly distributed random numbers.

    Parameters
    ----------
    shape : tuple or list
        The shape of the sampled random numbers.
    lower_bound : int or float
        The inclusive lower bound of the interval of the uniform distribution.
    upper_bound : int or float
        The exclusive upper bound of the interval of the uniform distribution.
    seed : int or None, optional
        A seed for the random number generator. Defaults to None,
        in which case a random value for the seed is used.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        A tensor containing a sample of uniformly distributed random numbers
        with shape ``shape``.

    See Also
    --------
    :func:`Graph.random.choices <random.choices>`
        Create random samples from the data that you provide.
    :func:`Graph.random.normal <random.normal>`
        Create a sample of normally distributed random numbers.
    :func:`boulderopal.run_stochastic_optimization`
        Function to find the minimum of generic stochastic functions.

    Examples
    --------
    Create a random tensor by sampling uniformly from :math:`[0,\, 1)`.

    >>> samples = graph.random.uniform(
    ...     shape=(3, 1), lower_bound=0, upper_bound=1, seed=0, name="samples"
    ... )
    >>> result = bo.execute_graph(graph=graph, output_node_names="samples")
    >>> result["output"]["samples"]["value"]
    array([[0.8069013], [0.79011373], [0.38818516]])

    See more examples in the `How to optimize controls robust to strong noise sources
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-
    robust-to-strong-noise-sources>`_ user guide.
    """

    name = "random_uniform"

    args = [
        forge.arg("shape", type=Union[Tuple, List]),
        forge.arg("lower_bound", type=Union[int, float]),
        forge.arg("upper_bound", type=Union[int, float]),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    rtype = Tensor
    categories = [Category.RANDOM_OPERATIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        shape = kwargs.get("shape")
        lower_bound = kwargs.get("lower_bound")
        upper_bound = kwargs.get("upper_bound")
        seed = kwargs.get("seed")

        check_argument(
            isinstance(shape, (tuple, list)),
            "The type of `shape` must be tuple or list.",
            {"shape": shape},
        )

        check_argument(
            lower_bound < upper_bound,
            "The lower bound must be smaller than the upper bound.",
            {"lower_bound": lower_bound, "upper_bound": upper_bound},
        )

        if seed is not None:
            check_argument_integer(seed, "seed")

        return Tensor(_operation, shape=tuple(shape))


class RandomChoices(Node):
    r"""
    Create random samples from the data that you provide.

    You can provide the data as a list and each element of that list represents one component
    of the full data. For example, considering a single variable linear regression problem
    that is described by the input :math:`x` and output :math:`y`, the data you provide would
    be :math:`[x, y]`. The first dimension of the data component in this list is the size of
    the data and therefore must be same for all components. However, all these components can
    have different value shapes, meaning the other dimensions can vary.

    This node effectively chooses a random batch of `sample_count` indices :math:`\{s_i\}`,
    and extracts the corresponding slices :math:`\{c[s_i]\}` of each data component.
    For example, in the case of linear regression, you can use this node to extract a random
    subset of your full data set.

    If this node is evaluated multiple times (for example during an optimization), it samples
    indices without replacement until all indices have been seen, at which point it starts sampling
    from the full set of indices again. You can therefore use this node to create minibatches that
    iterate over your data set in a series of epochs.

    Parameters
    ----------
    data : list[np.ndarray or Tensor]
        A list of data components. The first dimensions of the elements in this
        list denote the total amount of the data, and therefore must be the same.
    sample_count : int
        Number of samples in the returned batch.
    seed : int or None, optional
        Seed for random number generator. Defaults to None. If set, it ensures the
        random samples are generated in a reproducible sequence.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Sequence[Tensor]
        A sequence representing a batch of random samples from `data`.
        You can access the elements of the sequence using integer indices.
        The number of elements of the sequence is the same as the size of `data`.
        Each element of the sequence has the length (along its first dimension)
        as defined by `sample_count`.

    See Also
    --------
    :func:`Graph.random.normal <random.normal>`
        Create a sample of normally distributed random numbers.
    :func:`Graph.random.uniform <random.uniform>`
        Create a sample of uniformly distributed random numbers.
    :func:`boulderopal.run_stochastic_optimization`
        Function to find the minimum of generic stochastic functions.

    Examples
    --------
    >>> x = np.arange(20).reshape((10, 2))
    >>> y = np.arange(10) * 0.2
    >>> sampled_x, sampled_y = graph.random.choices([x, y], 3, seed=1)
    >>> sampled_x.name = "sampled_x"
    >>> sampled_y.name = "sampled_y"
    >>> result = bo.execute_graph(graph=graph, output_node_names=["sampled_x", "sampled_y"])
    >>> result["output"]["sampled_x"]["value"]
    array([[2, 3], [6, 7], [0, 1]])
    >>> result["output"]["sampled_y"]["value"]
    array([0.2, 0.6, 0. ])

    See more examples in the `How to perform Hamiltonian parameter estimation using a large
    amount of measured data
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-perform-parameter-estimation-
    with-a-large-amount-of-data>`_ user guide.
    """

    name = "random_choices"
    args = [
        forge.arg("data", type=List[TensorLike]),
        forge.arg("sample_count", type=int),
        forge.arg("seed", type=Optional[int], default=None),
    ]
    rtype = Sequence[Tensor]
    categories = [Category.RANDOM_OPERATIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        data = kwargs.get("data")
        sample_count = kwargs.get("sample_count")
        seed = kwargs.get("seed")

        data_size = data[0].shape[0]
        check_argument(
            all((value.shape[0] == data_size for value in data)),
            "The first dimension of the elements in `data` must be the same.",
            arguments={"data": data},
        )
        check_argument(
            0 < sample_count <= data_size,
            "`sample_count` must be positive and not greater than the size of the "
            "data you provide.",
            arguments={"sample_count": sample_count},
            extras={"data size": data_size},
        )
        if seed is not None:
            check_argument_integer(seed, "seed")

        return_tensor_shapes = [(sample_count,) + value.shape[1:] for value in data]

        return Sequence_(_operation).create_sequence(
            node_constructor=lambda operation, index: Tensor(
                operation, return_tensor_shapes[index]
            ),
            size=len(data),
        )


class RandomNamespace:
    """
    Operations to create random quantities.
    """

    __name__ = "random"

    def __init__(self, graph):
        self._graph = graph

    # pylint: disable=missing-function-docstring
    @wraps(RandomChoices.create_graph_method())
    def choices(self, *args, **kwargs):
        return RandomChoices.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(RandomColoredNoiseStfSignal.create_graph_method())
    def colored_noise_stf_signal(self, *args, **kwargs):
        return RandomColoredNoiseStfSignal.create_graph_method()(
            self._graph, *args, **kwargs
        )

    @wraps(RandomNormal.create_graph_method())
    def normal(self, *args, **kwargs):
        return RandomNormal.create_graph_method()(self._graph, *args, **kwargs)

    @wraps(RandomUniform.create_graph_method())
    def uniform(self, *args, **kwargs):
        return RandomUniform.create_graph_method()(self._graph, *args, **kwargs)


# Note that the names of all legacy random nodes all start with `random_`
# and we want to drop it in the new client.
RandomNamespaceDoc = create_namespace_doc(
    "random", __name__, [Category.RANDOM_OPERATIONS], lambda x: x[len("random_") :]
)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from functools import partial
from typing import Optional

import numpy as np

from boulderopal._nodes.documentation import Category
from boulderopal._nodes.node_data import Tensor
from boulderopal._validation.basic import ScalarDType

from .namespace import composite_node


@composite_node([Category.OPTIMIZATION_VARIABLES])
def optimizable_scalar(
    graph: "Graph",
    lower_bound: float,
    upper_bound: float,
    is_lower_unbounded: bool = False,
    is_upper_unbounded: bool = False,
    initial_values: Optional[float | list[float]] = None,
    name: Optional[str] = None,
) -> Tensor:
    r"""
    Create an optimizable scalar Tensor, which can be bounded, semi-bounded, or unbounded.

    Use this function to create a single variable that can be tuned by
    the optimizer to minimize the cost function.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    lower_bound : float
        The lower bound :math:`v_\mathrm{min}` for generating an initial value for the variable.
        This will also be used as lower bound if the variable is lower bounded.
    upper_bound : float
        The upper bound :math:`v_\mathrm{max}` for generating an initial value for the variable.
        This will also be used as upper bound if the variable is upper bounded.
    is_lower_unbounded : bool, optional
        Defaults to False. Set this flag to True to define a semi-bounded variable with
        lower bound :math:`-\infty`; in this case, the `lower_bound` parameter is used only for
        generating an initial value.
    is_upper_unbounded : bool, optional
        Defaults to False. Set this flag to True to define a semi-bounded variable with
        upper bound :math:`+\infty`; in this case, the `upper_bound` parameter is used only for
        generating an initial value.
    initial_values : float or List[float] or None, optional
        Initial values for the optimization variable. You can either provide a single initial
        value, or a list of them. Note that all optimization variables in a graph with non-default
        initial values must have the same length. That is, you must set them all either as a single
        value or a list of values of the same length. Defaults to None, meaning the optimizer
        initializes the variable with a random value.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The :math:`v` optimizable scalar. If both `is_lower_unbounded` and `is_upper_unbounded` are
        False, the variables is bounded such that :math:`v_\mathrm{min}\leq v \leq v_\mathrm{max}`.
        If one of the flags is True (for example `is_lower_unbounded=True`), the variable is
        semi-bounded (for example :math:`-\infty \leq v \leq v_\mathrm{max}`).
        If both of them are True, then the variable is unbounded and satisfies that
        :math:`-\infty \leq v \leq +\infty`.

    See Also
    --------
    Graph.optimization_variable : Create 1D Tensor of optimization variables.
    boulderopal.run_optimization :
        Function to find the minimum of a generic function.
    """

    if initial_values is not None:
        _validate_value = partial(
            ScalarDType.REAL,
            min_=lower_bound,
            max_=upper_bound,
            min_inclusive=True,
            max_inclusive=True,
        )
        if isinstance(initial_values, list):
            initial_values = [
                np.array([_validate_value(value, f"initial_values[{idx}]")])
                for idx, value in enumerate(initial_values)
            ]
        else:
            initial_values = np.array(
                [_validate_value(initial_values, "initial_values")]
            )

    scalar = graph.optimization_variable(
        count=1,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        is_lower_unbounded=is_lower_unbounded,
        is_upper_unbounded=is_upper_unbounded,
        initial_values=initial_values,
    )[0]

    if name is not None:
        scalar.name = name

    return scalar



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Functionality related to the computational-flow graph object.

The Graph object and its associated data types (Tensor, Pwc, ...) are re-imported here to
allow their access directly from the client package.
"""
from typing import Any

from numpydoc.docscrape import NumpyDocString
from qctrlcommons.graph import Graph as BaseGraph

from boulderopal._nodes.composite.registry import (
    COMPOSITE_NAMESPACE_REGISTRY,
    COMPOSITE_NODE_REGISTRY,
)
from boulderopal._nodes.registry import (
    PRIMARY_NAMESPACE_REGISTRY,
    PRIMARY_NODE_REGISTRY,
)

# List of nodes that won't be migrated to the new Boulder Opal Client.
_OBSOLETE_NODES = [
    "random_choices",
    "random_colored_noise_stf_signal",
    "random_normal",
    "random_uniform",
]
_OBSOLETE_NODES += [
    "ms_dephasing_robust_cost",
    "ms_displacements",
    "ms_infidelity",
    "ms_phases",
    "ms_phases_multitone",
]


def _extend_method(
    obj: Any, method: str, method_name: str, check_exists: bool = True
) -> None:
    """
    Extend the specified object by adding methods as attributes.

    Parameters
    ----------
    obj : Any
        The object to which the node should be added.
    method : Any
        Method to be added to the object.
    method_name : str
        Name of the method to be added.
    check_exists : bool
        Whether to check if the method already exists as an
        attribute on the object.
    """
    if check_exists and hasattr(obj, method_name):
        raise AttributeError(f"existing attr {method_name} on namespace: {obj}")
    setattr(obj, method_name, method)


def _clean_doc(doc: str) -> str:
    """
    Remove the graph parameter from the doc.
    """
    doc_obj = NumpyDocString(doc)

    doc_obj["Parameters"] = [
        item for item in doc_obj["Parameters"] if item.name != "graph"
    ]

    return str(doc_obj)


class Graph(BaseGraph):  # pylint: disable=too-few-public-methods
    """
    A class for representing and building a Boulder Opal data flow graph.

    The graph object is the main entry point to the Boulder Opal graph ecosystem.
    You can call methods to add nodes to the graph, and use the `operations` attribute to get a
    dictionary representation of the graph.
    """

    def __init__(self) -> None:
        self._add_namespaces()
        super().__init__()

    def _add_namespaces(self) -> None:
        # We initialize the node namespaces to the graph in this way since they need access
        # to the initialized class object.
        for _composite_ns in COMPOSITE_NAMESPACE_REGISTRY:
            _extend_method(self, _composite_ns(self), _composite_ns.namespace_name)

        for _primary_ns in PRIMARY_NAMESPACE_REGISTRY:
            _ns = _primary_ns(self)
            _extend_method(self, _ns, _ns.__name__)  # type: ignore


# Set nodes to Graph.
for node_cls in PRIMARY_NODE_REGISTRY.as_list():
    node = node_cls.create_graph_method()
    if node.__name__ not in _OBSOLETE_NODES:
        _extend_method(Graph, node, node.__name__)

# Set composite nodes to Graph.
for composite_node in COMPOSITE_NODE_REGISTRY:
    composite_node.__doc__ = _clean_doc(composite_node.__doc__)
    _extend_method(Graph, composite_node, composite_node.__name__)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.node_data import Pwc
from boulderopal._validation.basic import ScalarDType

if TYPE_CHECKING:
    from boulderopal.graph.graph import Graph


class OptimizableCoefficient(ABC):
    """
    Abstract class for optimizable Hamiltonian coefficients.
    """

    @abstractmethod
    def get_pwc(self, graph: Graph, gate_duration: float, name: str) -> Pwc:
        """
        Return a Pwc representation of the optimizable coefficient.
        """
        raise NotImplementedError


@dataclass
class RealOptimizableConstant(OptimizableCoefficient):
    """
    A real-valued optimizable constant coefficient for a Hamiltonian term.
    The main function will try to find the optimal value for this constant.

    Parameters
    ----------
    min : float
        The minimum value that the coefficient can take.
    max : float
        The maximum value that the coefficient can take.

    See Also
    --------
    boulderopal.superconducting.ComplexOptimizableConstant :
        Class describing complex optimizable constant coefficients.
    boulderopal.superconducting.ComplexOptimizableSignal :
        Class describing complex optimizable piecewise-constant coefficients.
    boulderopal.superconducting.RealOptimizableSignal :
        Class describing real optimizable piecewise-constant coefficients.
    """

    min: float
    max: float

    def __post_init__(self) -> None:
        self.min = ScalarDType.REAL(self.min, "min")
        self.max = ScalarDType.REAL(self.max, "max")
        check_argument(
            self.min < self.max,
            "The maximum must be larger than the minimum.",
            {"min": self.min, "max": self.max},
        )

    def get_pwc(self, graph: Graph, gate_duration: float, name: str) -> Pwc:
        value = graph.optimizable_scalar(self.min, self.max)
        value.name = name
        return graph.constant_pwc(constant=value, duration=gate_duration)


@dataclass
class ComplexOptimizableConstant(OptimizableCoefficient):
    """
    A complex-valued optimizable constant coefficient for a Hamiltonian term.
    The main function will try to find the optimal value for this constant.

    Parameters
    ----------
    max_modulus : float
        The maximum value that the modulus of the coefficient can take.

    See Also
    --------
    boulderopal.superconducting.ComplexOptimizableSignal :
        Class describing complex optimizable piecewise-constant coefficients.
    boulderopal.superconducting.RealOptimizableConstant :
        Class describing real optimizable constant coefficients.
    boulderopal.superconducting.RealOptimizableSignal :
        Class describing real optimizable piecewise-constant coefficients.
    """

    max_modulus: float

    def __post_init__(self) -> None:
        self.max_modulus = ScalarDType.REAL(self.max_modulus, "max_modulus", min_=0)

    def get_pwc(self, graph: Graph, gate_duration: float, name: str) -> Pwc:
        mod = graph.optimizable_scalar(0, self.max_modulus)
        phase = graph.optimizable_scalar(0, 2 * np.pi, True, True)
        value = graph.multiply(mod, graph.exp(1j * phase), name=name)
        return graph.constant_pwc(constant=value, duration=gate_duration)


@dataclass
class RealOptimizableSignal(OptimizableCoefficient):
    """
    A real-valued optimizable time-dependent piecewise-constant coefficient for
    a Hamiltonian term. The main function will try to find the optimal value for
    this signal at each segment.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant signal.
    min : float
        The minimum value that the signal can take at each segment.
    max : float
        The maximum value that the signal can take at each segment.

    See Also
    --------
    boulderopal.superconducting.ComplexOptimizableConstant :
        Class describing complex optimizable constant coefficient.
    boulderopal.superconducting.ComplexOptimizableSignal :
        Class describing complex optimizable piecewise-constant coefficients.
    boulderopal.superconducting.RealOptimizableConstant :
        Class describing real optimizable constant coefficients.
    """

    count: int
    min: float
    max: float

    def __post_init__(self) -> None:
        self.count = ScalarDType.INT(self.count, "count", min_=0)
        self.min = ScalarDType.REAL(self.min, "min")
        self.max = ScalarDType.REAL(self.max, "max")
        check_argument(
            self.min < self.max,
            "The maximum must be larger than the minimum.",
            {"min": self.min, "max": self.max},
        )

    def get_pwc(self, graph: Graph, gate_duration: float, name: str) -> Pwc:
        values = graph.optimization_variable(self.count, self.min, self.max)
        return graph.pwc_signal(values=values, duration=gate_duration, name=name)


@dataclass
class ComplexOptimizableSignal(OptimizableCoefficient):
    """
    A complex-valued optimizable time-dependent piecewise-constant coefficient
    for a Hamiltonian term. The main function will try to find the optimal value
    for this signal at each segment.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant signal.
    max_modulus : float
        The maximum value that the modulus of the signal can take at each segment.

    See Also
    --------
    boulderopal.superconducting.ComplexOptimizableConstant :
        Class describing complex optimizable constant coefficients.
    boulderopal.superconducting.RealOptimizableConstant :
        Class describing real optimizable constant coefficients.
    boulderopal.superconducting.RealOptimizableSignal :
        Class describing real optimizable piecewise-constant coefficients.
    """

    count: int
    max_modulus: float

    def __post_init__(self) -> None:
        self.count = ScalarDType.INT(self.count, "count", min_=0)
        self.max_modulus = ScalarDType.REAL(self.max_modulus, "max_modulus", min_=0)

    def get_pwc(self, graph: Graph, gate_duration: float, name: str) -> Pwc:
        return graph.complex_optimizable_pwc_signal(
            segment_count=self.count,
            duration=gate_duration,
            maximum=self.max_modulus,
            name=name,
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_duration,
)

from . import node_data
from .documentation import Category
from .utils import (
    check_optimization_variable_parameters,
    validate_inputs_real_fourier_signal,
)


class OptimizationVariable(Node):
    r"""
    Create a 1D Tensor of optimization variables, which can be bounded,
    semi-bounded, or unbounded.

    Use this function to create a sequence of variables that can be tuned by
    the optimizer (within specified bounds) in order to minimize the cost
    function.

    Parameters
    ----------
    count : int
        The number :math:`N` of individual real-valued variables to create.
    lower_bound : float
        The lower bound :math:`v_\mathrm{min}` for generating an initial value for the variables.
        This will also be used as lower bound if the variables are lower bounded.
        The same lower bound applies to all `count` individual variables.
    upper_bound : float
        The upper bound :math:`v_\mathrm{max}` for generating an initial value for the variables.
        This will also be used as upper bound if the variables are upper bounded.
        The same upper bound applies to all `count` individual variables.
    is_lower_unbounded : bool, optional
        Defaults to False. Set this flag to True to define a semi-bounded variable with
        lower bound :math:`-\infty`; in this case, the `lower_bound` parameter is used only for
        generating an initial value.
    is_upper_unbounded : bool, optional
        Defaults to False. Set this flag to True to define a semi-bounded variable with
        upper bound :math:`+\infty`; in this case, the `upper_bound` parameter is used only for
        generating an initial value.
    initial_values : np.ndarray or List[np.ndarray] or None, optional
        Initial values for the optimization variable. You can either provide a single initial
        value, or a list of them. Note that all optimization variables in a graph with non-default
        initial values must have the same length. That is, you must set them either as a single
        array or a list of arrays of the same length. Defaults to None, meaning the optimizer
        initializes the variables with random values.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The sequence :math:`\{v_n\}` of :math:`N` optimization variables. If both
        `is_lower_unbounded` and `is_upper_unbounded` are False, these variables are
        bounded such that :math:`v_\mathrm{min}\leq v_n\leq v_\mathrm{max}`. If one of the
        flags is True (for example `is_lower_unbounded=True`), these variables are
        semi-bounded (for example :math:`-\infty \leq v_n \leq v_\mathrm{max}`).
        If both of them are True, then these variables are unbounded and satisfy that
        :math:`-\infty \leq v_n \leq +\infty`.

    See Also
    --------
    Graph.anchored_difference_bounded_variables :
        Create anchored optimization variables with a difference bound.
    Graph.complex_optimizable_pwc_signal :
        Create a complex optimizable `Pwc` signal.
    Graph.optimizable_scalar : Create an optimization scalar.
    Graph.real_optimizable_pwc_signal :
        Create a real optimizable `Pwc` signal.
    boulderopal.run_optimization :
        Function to find the minimum of a generic function.

    Examples
    --------
    Perform a simple optimization task.

    >>> variables = graph.optimization_variable(
    ...     2, lower_bound=0, upper_bound=1, name="variables"
    ... )
    >>> x = variables[0]
    >>> y = variables[1]
    >>> cost = (x - 0.1) ** 2 + graph.sin(y) ** 2
    >>> cost.name = "cost"
    >>> result = bo.run_optimization(
    ...     graph=graph, cost_node_name="cost", output_node_names="variables"
    ... )
    >>> result["cost"]
    0.0
    >>> result["output"]["variables"]["value"]
    array([0.1, 0.])

    See examples about optimal control of quantum systems in the
    `How to optimize controls in arbitrary quantum systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-
    in-arbitrary-quantum-systems-using-graphs>`_ user guide.
    """

    name = "optimization_variable"
    optimizable_variable = True
    args = [
        forge.arg("count", type=int),
        forge.arg("lower_bound", type=float),
        forge.arg("upper_bound", type=float),
        forge.arg("is_lower_unbounded", type=bool, default=False),
        forge.arg("is_upper_unbounded", type=bool, default=False),
        forge.arg(
            "initial_values",
            type=Optional[Union[np.ndarray, List[np.ndarray]]],
            default=None,
        ),
    ]
    rtype = node_data.Tensor
    categories = [Category.OPTIMIZATION_VARIABLES]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        count = kwargs["count"]
        lower_bound = kwargs["lower_bound"]
        upper_bound = kwargs["upper_bound"]
        initial_values = kwargs["initial_values"]
        check_optimization_variable_parameters(
            initial_values, count, lower_bound, upper_bound
        )
        return node_data.Tensor(_operation, shape=(count,))


class AnchoredDifferenceBoundedVariables(Node):
    r"""
    Create a sequence of optimizable variables with an anchored difference bound.

    Use this function to create a sequence of optimization variables that have
    a difference bound (each variable is constrained to be within a given
    distance of the adjacent variables) and are anchored to zero at the start
    and end (the initial and final variables are within a given distance of
    zero).

    Parameters
    ----------
    count : int
        The number :math:`N` of individual real-valued variables to create.
    lower_bound : float
        The lower bound :math:`v_\mathrm{min}` on the variables.
        The same lower bound applies to all `count` individual variables.
    upper_bound : float
        The upper bound :math:`v_\mathrm{max}` on the variables.
        The same upper bound applies to all `count` individual variables.
    difference_bound : float
        The difference bound :math:`\delta` to enforce between adjacent
        variables.
    initial_values : np.ndarray or List[np.ndarray] or None, optional
        Initial values for optimization variable. You can either provide a single initial
        value, or a list of them. Note that all optimization variables in a graph with non-default
        initial values must have the same length. That is, you must set them either as a single
        array or a list of arrays of the same length. Defaults to None, meaning the optimizer
        initializes the variables with random values.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Tensor
        The sequence :math:`\{v_n\}` of :math:`N` anchored difference-bounded
        optimization variables, satisfying
        :math:`v_\mathrm{min}\leq v_n\leq v_\mathrm{max}`,
        :math:`|v_{n-1}-v_n|\leq\delta` for :math:`2\leq n\leq N`,
        :math:`|v_1|\leq\delta`, and :math:`|v_N|\leq\delta`.

    See Also
    --------
    Graph.optimization_variable : Create 1D Tensor of optimization variables.
    boulderopal.run_optimization :
        Function to find the minimum of generic deterministic functions.
    boulderopal.run_stochastic_optimization :
        Function to find the minimum of generic stochastic functions.

    Examples
    --------
    Create optimizable PWC signal with anchored difference bound.

    >>> values = graph.anchored_difference_bounded_variables(
    ...     count=10, lower_bound=-1, upper_bound=1, difference_bound=0.1
    ... )
    >>> graph.pwc_signal(values=values, duration=1)
    <Pwc: name="pwc_signal_#1", operation_name="pwc_signal", value_shape=(), batch_shape=()>

    See the "Band-limited pulses with bounded slew rates" example in the
    `How to add smoothing and band-limits to optimized controls
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-add-smoothing-and-band-limits-to-
    optimized-controls#example-band-limited-pulses-with-bounded-slew-rates>`_ user guide.
    """

    name = "anchored_difference_bounded_variables"
    optimizable_variable = True
    args = [
        forge.arg("count", type=int),
        forge.arg("lower_bound", type=float),
        forge.arg("upper_bound", type=float),
        forge.arg("difference_bound", type=float),
        forge.arg(
            "initial_values",
            type=Optional[Union[np.ndarray, List[np.ndarray]]],
            default=None,
        ),
    ]
    rtype = node_data.Tensor
    categories = [Category.OPTIMIZATION_VARIABLES]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        count = kwargs["count"]
        lower_bound = kwargs["lower_bound"]
        upper_bound = kwargs["upper_bound"]
        initial_values = kwargs["initial_values"]
        check_optimization_variable_parameters(
            initial_values, count, lower_bound, upper_bound
        )
        return node_data.Tensor(_operation, shape=(count,))


class RealFourierStfSignal(Node):
    r"""
    Create a real sampleable signal constructed from Fourier components.

    Use this function to create a signal defined in terms of Fourier (sine/cosine)
    basis signals that can be optimized by varying their coefficients and, optionally,
    their frequencies.

    Parameters
    ----------
    duration : float
        The total duration :math:`\tau` of the signal.
    initial_coefficient_lower_bound : float, optional
        The lower bound :math:`c_\mathrm{min}` on the initial coefficient
        values. Defaults to -1.
    initial_coefficient_upper_bound : float, optional
        The upper bound :math:`c_\mathrm{max}` on the initial coefficient
        values. Defaults to 1.
    fixed_frequencies : list[float] or None, optional
        The fixed non-zero frequencies :math:`\{f_m\}` to use for the Fourier
        basis. If provided, must be non-empty and specified in the inverse
        units of `duration` (for example if `duration` is in seconds, these
        values must be given in Hertz).
    optimizable_frequency_count : int or None, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies can be optimized. Defaults to 0.
    randomized_frequency_count : int or None, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies are to be randomized but fixed. Defaults to 0.

    Returns
    -------
    Stf(1D, real)
        The optimizable, real-valued, sampleable signal built from the
        appropriate Fourier components.

    Warnings
    --------
    You must provide exactly one of `fixed_frequencies`, `optimizable_variable`,
    or `randomized_frequency_count`.

    See Also
    --------
    Graph.real_fourier_pwc_signal : Corresponding operation for `Pwc`.

    Notes
    -----
    This function sets the basis signal frequencies :math:`\{f_m\}`
    depending on the chosen mode:

    * For fixed frequencies, you provide the frequencies directly.
    * For optimizable frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` unbounded optimizable
      variables :math:`\{f_m\}`, with initial values in the ranges
      :math:`\{[(m-1)/\tau, m/\tau]\}`.
    * For randomized frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` randomized constants
      :math:`\{f_m\}` in the ranges :math:`\{[(m-1)/\tau, m/\tau]\}`.

    After this function creates the :math:`M` frequencies :math:`\{f_m\}`, it
    produces the signal

    .. math::
        \alpha^\prime(t) = v_0 +
        \sum_{m=1}^M [ v_m \cos(2\pi t f_m) + w_m \sin(2\pi t f_m) ],

    where :math:`\{v_m,w_m\}` are (unbounded) optimizable variables, with
    initial values bounded by :math:`c_\mathrm{min}` and
    :math:`c_\mathrm{max}`. This function produces the final signal :math:`\alpha(t)`.

    You can use the signals created by this function for chopped random basis
    (CRAB) optimization [1]_.

    References
    ----------
    .. [1] `P. Doria, T. Calarco, and S. Montangero,
            Phys. Rev. Lett. 106, 190501 (2011).
            <https://doi.org/10.1103/PhysRevLett.106.190501>`_
    """

    name = "real_fourier_stf_signal"
    optimizable_variable = True
    args = [
        forge.arg("duration", type=float),
        forge.arg("initial_coefficient_lower_bound", type=float, default=-1),
        forge.arg("initial_coefficient_upper_bound", type=float, default=1),
        forge.arg("fixed_frequencies", type=Optional[List[float]], default=None),
        forge.arg("optimizable_frequency_count", type=Optional[int], default=None),
        forge.arg("randomized_frequency_count", type=Optional[int], default=None),
    ]
    kwargs = {}  # Stfs don't accept name as an argument.
    rtype = node_data.Stf
    categories = [
        Category.BUILDING_SMOOTH_HAMILTONIANS,
        Category.OPTIMIZATION_VARIABLES,
    ]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        duration = kwargs.get("duration")
        check_duration(duration, "duration")

        fixed_frequencies = kwargs.get("fixed_frequencies")
        optimizable_frequency_count = kwargs.get("optimizable_frequency_count")
        randomized_frequency_count = kwargs.get("randomized_frequency_count")
        validate_inputs_real_fourier_signal(
            fixed_frequencies, optimizable_frequency_count, randomized_frequency_count
        )

        return node_data.Stf(_operation, value_shape=(), batch_shape=())


class RealFourierPwcSignal(Node):
    r"""
    Create a piecewise-constant signal constructed from Fourier components.

    Use this function to create a signal defined in terms of Fourier
    (sine/cosine) basis signals that can be optimized by varying their
    coefficients and, optionally, their frequencies.

    Parameters
    ----------
    duration : float
        The total duration :math:`\tau` of the signal.
    segment_count : int
        The number of segments :math:`N` to use for the piecewise-constant
        approximation.
    initial_coefficient_lower_bound : float, optional
        The lower bound :math:`c_\mathrm{min}` on the initial coefficient
        values. Defaults to -1.
    initial_coefficient_upper_bound : float, optional
        The upper bound :math:`c_\mathrm{max}` on the initial coefficient
        values. Defaults to 1.
    fixed_frequencies : list[float] or None, optional
        The fixed non-zero frequencies :math:`\{f_m\}` to use for the Fourier
        basis. If provided, must be non-empty and specified in the inverse
        units of `duration` (for example if `duration` is in seconds, these
        values must be given in Hertz).
    optimizable_frequency_count : int or None, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies can be optimized. Defaults to 0.
    randomized_frequency_count : int or None, optional
        The number of non-zero frequencies :math:`M` to use, if the
        frequencies are to be randomized but fixed. Defaults to 0.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Pwc(1D, real)
        The optimizable, real-valued, piecewise-constant signal built from the
        appropriate Fourier components.

    Warnings
    --------
    You must provide exactly one of `fixed_frequencies`, `optimizable_variable`,
    or `randomized_frequency_count`.

    See Also
    --------
    Graph.real_fourier_stf_signal : Corresponding operation for `Stf`.

    Notes
    -----
    This function sets the basis signal frequencies :math:`\{f_m\}`
    depending on the chosen mode:

    * For fixed frequencies, you provide the frequencies directly.
    * For optimizable frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` unbounded optimizable
      variables :math:`\{f_m\}`, with initial values in the ranges
      :math:`\{[(m-1)/\tau, m/\tau]\}`.
    * For randomized frequencies, you provide the number of frequencies
      :math:`M`, and this function creates :math:`M` randomized constants
      :math:`\{f_m\}` in the ranges :math:`\{[(m-1)/\tau, m/\tau]\}`.

    After this function creates the :math:`M` frequencies :math:`\{f_m\}`, it
    produces the signal

    .. math::
        \alpha^\prime(t) = v_0 +
        \sum_{m=1}^M [ v_m \cos(2\pi t f_m) + w_m \sin(2\pi t f_m) ],

    where :math:`\{v_m,w_m\}` are (unbounded) optimizable variables, with
    initial values bounded by :math:`c_\mathrm{min}` and
    :math:`c_\mathrm{max}`. This function produces the final
    piecewise-constant signal :math:`\alpha(t)` by sampling
    :math:`\alpha^\prime(t)` at :math:`N` equally spaced points along the
    duration :math:`\tau`, and using those sampled values as the constant
    segment values.

    You can use the signals created by this function for chopped random basis
    (CRAB) optimization [1]_.

    References
    ----------
    .. [1] `P. Doria, T. Calarco, and S. Montangero,
            Phys. Rev. Lett. 106, 190501 (2011).
            <https://doi.org/10.1103/PhysRevLett.106.190501>`_

    Examples
    --------
    See the "CRAB optimization on a qutrit" example in the
    `How to perform model-based optimization using a Fourier basis
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-using-
    a-fourier-basis#example-crab-optimization-on-a-qutrit>`_ user guide.
    """

    name = "real_fourier_pwc_signal"
    optimizable_variable = True
    args = [
        forge.arg("duration", type=float),
        forge.arg("segment_count", type=int),
        forge.arg("initial_coefficient_lower_bound", type=float, default=-1),
        forge.arg("initial_coefficient_upper_bound", type=float, default=1),
        forge.arg("fixed_frequencies", type=Optional[List[float]], default=None),
        forge.arg("optimizable_frequency_count", type=Optional[int], default=None),
        forge.arg("randomized_frequency_count", type=Optional[int], default=None),
    ]
    rtype = node_data.Pwc
    categories = [
        Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS,
        Category.OPTIMIZATION_VARIABLES,
    ]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        duration = kwargs.get("duration")
        segment_count = kwargs.get("segment_count")
        check_duration(duration, "duration")

        check_argument_integer(segment_count, "segment_count")
        check_argument(
            segment_count > 0,
            "The number of segments must be greater than zero.",
            {"segment_count": segment_count},
        )

        fixed_frequencies = kwargs.get("fixed_frequencies")
        optimizable_frequency_count = kwargs.get("optimizable_frequency_count")
        randomized_frequency_count = kwargs.get("randomized_frequency_count")
        validate_inputs_real_fourier_signal(
            fixed_frequencies, optimizable_frequency_count, randomized_frequency_count
        )

        durations = duration / segment_count * np.ones(segment_count)

        return node_data.Pwc(
            _operation, value_shape=(), durations=durations, batch_shape=()
        )
